#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot._compat import TypedDict
from datarobot.models.api_object import APIObject
from datarobot.models.notebooks.enums import (
    CellType,
    KernelExecutionStatus,
    NotebookStatus,
    SessionType,
)

# TODO: We're using trafaret's "ignore_extra" liberally and this is a subset of properties
notebook_session_trafaret = t.Dict({
    t.Key("status"): t.Enum(*list(NotebookStatus)),
    t.Key("notebook_id"): t.String,
    t.Key("session_id"): t.String,
    t.Key("started_at", optional=True): t.String,
    t.Key("session_type", optional=True): t.Enum(*list(SessionType)),
    t.Key("ephemeral_session_key", optional=True): t.String,
}).ignore_extra("*")


# TODO: We're using trafaret's "ignore_extra" - this is a subset of properties
notebook_execution_status_trafaret = t.Dict({
    t.Key("status"): t.Enum(*list(KernelExecutionStatus)),
    t.Key("cell_id", optional=True): t.String,
    t.Key("queued_cell_ids", optional=True): t.List(t.String),
}).ignore_extra("*")


codespace_notebook_cell_trafaret = t.Dict({
    t.Key("id"): t.String,
    t.Key("cell_type"): t.Enum(*list(CellType)),
    t.Key("source"): t.String(allow_blank=True) | t.List(t.String(allow_blank=True)),
    t.Key("metadata"): t.Dict({}).allow_extra("*"),  # TODO: Better annotate this
    t.Key("execution_count", optional=True): t.Int,
}).ignore_extra("*")


codespace_notebook_state_trafaret = t.Dict({
    t.Key("name"): t.String,
    t.Key("path"): t.String,
    t.Key("generation"): t.Int,
    t.Key("nbformat"): t.Int,
    t.Key("nbformat_minor"): t.Int,
    t.Key("metadata"): t.Dict({}).allow_extra("*"),  # TODO: Better annotate this
    t.Key("cells"): t.List(codespace_notebook_cell_trafaret),
    t.Key("kernel_id", optional=True): t.String,
}).ignore_extra("*")


class CloneRepositorySchema(TypedDict):
    """
    Schema for cloning a repository when starting a notebook session.
    """

    url: str
    checkout_ref: Optional[str]


class StartSessionParameters(TypedDict):
    """
    Parameters used as environment variables in a notebook session.
    """

    name: str
    value: str


class StartSessionPayload(TypedDict, total=False):
    """
    Payload for starting a notebook session.
    """

    is_triggered_run: bool
    parameters: Optional[List[StartSessionParameters]]
    open_file_paths: Optional[List[str]]
    clone_repository: Optional[CloneRepositorySchema]


class NotebookExecutionStatus(APIObject):
    """
    Notebook execution status information.

    Attributes
    ----------

    status : str
        The status of the notebook execution.
    cell_id : Optional[bson.ObjectId]
        The ID of the cell being executed. Optional.
    queued_cell_ids : Optional[List[bson.ObjectId]]
        The list of cell IDs that are queued for execution. Optional.
    """

    _converter = notebook_execution_status_trafaret

    def __init__(
        self,
        status: KernelExecutionStatus,
        cell_id: Optional[str] = None,
        queued_cell_ids: Optional[str] = None,
    ):
        self.status = status
        self.cell_id = cell_id
        self.queued_cell_ids = queued_cell_ids


class CodespaceNotebookCell(TypedDict):
    """
    Represents a cell in a codespace notebook.
    """

    id: str
    cell_type: str
    source: str
    metadata: Dict[str, Any]  # TODO: Better annotate this
    execution_count: Optional[int]


class CodespaceNotebookState(APIObject):
    """
    Notebook state information for a codespace notebook.

    Attributes
    ----------
    name : str
        The name of the notebook.
    path : str
        The path of the notebook.
    generation : int
        The generation of the notebook.
    nbformat : int
        The notebook format version.
    nbformat_minor : int
        The notebook format minor version.
    metadata : dict
        The metadata of the notebook.
    cells : List[CodespaceNotebookCell]
        The list of cells in the notebook.
    kernel_id : Optional[str]
        The ID of the kernel. Optional.
    """

    _converter = codespace_notebook_state_trafaret

    def __init__(
        self,
        name: str,
        path: str,
        generation: int,
        nbformat: int,
        nbformat_minor: int,
        metadata: Dict[str, Any],  # TODO: Better annotate this
        cells: List[CodespaceNotebookCell],
        kernel_id: Optional[str] = None,
    ):
        self.name = name
        self.path = path
        self.generation = generation
        self.nbformat = nbformat
        self.nbformat_minor = nbformat_minor
        self.metadata = metadata
        self.cells = cells
        self.kernel_id = kernel_id


class NotebookSession(APIObject):
    """
    Notebook session information.

    Attributes
    ----------

    status : NotebookStatus
        The current status of the notebook kernel.
    notebook_id : str
        The ID of the notebook.
    session_id : str
        The ID of the session. Incorporates the ``notebook_id`` as part of this ID.
    started_at : Optional[str]
        The date and time when the notebook was started. Optional.
    session_type: Optional[SessionType]
        The type of the run - either manual (triggered via UI or API) or scheduled. Optional.
    ephemeral_session_key: Optional[str]
        The ID specific to ephemeral session if being used. Optional.
    """

    _runtimes_path = "notebookRuntimes/"
    _sessions_path = "notebookSessions/"

    _converter = notebook_session_trafaret

    def __init__(
        self,
        status: NotebookStatus,
        notebook_id: str,
        session_id: str,
        started_at: Optional[str] = None,
        session_type: Optional[SessionType] = None,
        ephemeral_session_key: Optional[str] = None,
    ):
        self.status = status
        self.notebook_id = notebook_id
        self.session_id = session_id
        self.started_at = started_at
        self.session_type = session_type
        self.ephemeral_session_key = ephemeral_session_key

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(status={self.status}, id={self.session_id}, notebook_id={self.notebook_id}, "
            f"type={self.session_type}, notebook_id={self.notebook_id})"
        )

    @classmethod
    def get(cls, notebook_id: str) -> NotebookSession:
        """
        Get a notebook session by its notebook ID.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook.

        Returns
        -------
        NotebookSession
            The notebook session information.
        """
        r_data = cls._client.get(f"{cls._runtimes_path}notebooks/{notebook_id}/")
        return cls.from_server_data(r_data.json())

    @classmethod
    def start(cls, notebook_id: str, payload: StartSessionPayload) -> NotebookSession:
        """
        Start a notebook session.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook.
        payload : StartSessionPayload
            The payload to start the session.

        Returns
        -------
        NotebookSession
            The notebook session information.
        """
        r_data = cls._client.post(f"{cls._runtimes_path}notebooks/{notebook_id}/start/", data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def stop(cls, notebook_id: str) -> NotebookSession:
        """
        Stop a notebook session.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook.

        Returns
        -------
        NotebookSession
            The notebook session information.
        """
        r_data = cls._client.post(f"{cls._runtimes_path}notebooks/{notebook_id}/stop/")
        return cls.from_server_data(r_data.json())

    @classmethod
    def execute_notebook(cls, notebook_id: str, cell_ids: Optional[List[str]] = None) -> None:
        """
        Execute a notebook.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook.
        cell_ids : Optional[List[bson.ObjectId]]
            The list of cell IDs to execute. Optional. If not provided, the whole notebook will be executed.
        """
        payload = {"cell_ids": cell_ids} if cell_ids else {}
        cls._client.post(f"{cls._runtimes_path}notebooks/{notebook_id}/execute/", data=payload)

    @classmethod
    def get_codespace_notebook_state(cls, notebook_id: str, notebook_path: str) -> CodespaceNotebookState:
        r_data = cls._client.get(f"{cls._sessions_path}{notebook_id}/notebook/", params={"path": notebook_path})
        return CodespaceNotebookState.from_server_data(r_data.json())

    @classmethod
    def execute_codespace_notebook(
        cls,
        notebook_id: str,
        notebook_path: str,
        generation: int,
        cells: List[CodespaceNotebookCell],
    ) -> None:
        """
        Execute a notebook.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook.
        notebook_path : str
            The path of the notebook.
        generation : int
            The generation of the notebook.
        cells : List[CodespaceNotebookCell]
            The list of cells to execute.
        """
        payload = {
            "path": notebook_path,
            "generation": generation,
            "cells": cells,
        }
        cls._client.post(f"{cls._sessions_path}{notebook_id}/notebook/execute/", data=payload)

    @classmethod
    def get_execution_status(cls, notebook_id: str) -> NotebookExecutionStatus:
        """
        Get the execution status information of a notebook.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook.

        Returns
        -------
        NotebookExecutionStatus
            The execution status information of the notebook.
        """
        r_data = cls._client.get(f"{cls._runtimes_path}notebooks/{notebook_id}/executionStatus/")
        return NotebookExecutionStatus.from_server_data(r_data.json())



#
# Copyright 2023-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""

  AbstractJob
  ^        ^
  |        |
  +        +
Job       AbstractSpecificJob
            ^             ^
            |             |
            +             +
        ModelJob   AbstractBatchJob
                    ^           ^
                    |           |
                    +           +
        BatchPredictionJob    BatchMonitoringJob


"""

from __future__ import annotations

import csv
import io
import os
import threading
import time
import typing
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union

import pandas as pd
import requests
import trafaret as t

from datarobot._compat import Int, String, TypedDict
from datarobot.utils import recognize_sourcedata as orig_recognize_sourcedata

from .. import errors
from ..enums import AVAILABLE_STATEMENT_TYPES, IntakeAdapters, OutputAdapters, TrainingDataSubsets
from .dataset import Dataset
from .job import AbstractSpecificJob


class IntakeSettings(TypedDict, total=False):
    """Intake settings typed dict"""

    type: IntakeAdapters
    file: Optional[Union[str, pd.DataFrame, io.IOBase]]
    url: Optional[str]
    credential_id: Optional[str]
    data_store_id: Optional[str]
    query: Optional[str]
    table: Optional[str]
    schema: Optional[str]
    catalog: Optional[str]
    fetch_size: Optional[int]
    format: Optional[str]
    endpoint_url: Optional[str]
    dataset: Optional[Dataset]
    dataset_id: Optional[str]
    dataset_version_id: Optional[str]


class OutputSettings(TypedDict, total=False):
    """Output settings typed dict"""

    type: OutputAdapters
    path: Optional[str]
    url: Optional[str]
    credential_id: Optional[str]
    data_store_id: Optional[str]
    table: Optional[str]
    schema: Optional[str]
    catalog: Optional[str]
    statement_type: Optional[str]
    update_columns: Optional[List[str]]
    where_columns: Optional[List[str]]
    create_table_if_not_exists: Optional[bool]


class CsvSettings(TypedDict):
    delimiter: Optional[str]
    quotechar: Optional[str]
    encoding: Optional[str]


class Schedule(TypedDict):
    day_of_week: List[Union[int, str]]
    month: List[Union[int, str]]
    hour: List[Union[int, str]]
    minute: List[Union[int, str]]
    day_of_month: List[Union[int, str]]


def recognize_sourcedata(sourcedata: Any, default_fname: str) -> Dict[str, Any]:
    """Override the default recognize_sourcedata with one that converts
    DataFrame to io.BytesIO"""
    if not isinstance(sourcedata, pd.DataFrame):
        return orig_recognize_sourcedata(sourcedata, default_fname)
    # The original recognize_sourcedata will encode dataframes into StringIO
    # To pass this to requests.put, we need BytesIO - otherwise, requests will
    # try to encode it using latin-1 encoding (the default HTTP encoding),
    # which will crash when a charater not contained in latin-1 is used
    #
    # https://github.com/python/cpython/blob/b9797417315cc2d1700cb2d427685016d3380711/Lib/http/client.py#L1046
    buf = io.BytesIO()
    sourcedata.to_csv(buf, encoding="utf-8", index=False, quoting=csv.QUOTE_ALL)
    buf.seek(0)
    return {"filelike": buf, "fname": default_fname}


class AbstractBatchJob(AbstractSpecificJob):
    """
    Generic representation of batch job, that covers common functions/checks
    for predictions, and monitoring jobs.

    """

    _s3_settings = t.Dict({
        t.Key("url"): String(),
        t.Key("credential_id", optional=True): String(),
        t.Key("endpoint_url", optional=True): String(),
        t.Key("format", optional=True): String(),
    })

    _gcp_settings = t.Dict({
        t.Key("url"): String(),
        t.Key("credential_id", optional=True): String(),
        t.Key("format", optional=True): String(),
    })

    _azure_settings = t.Dict({
        t.Key("url"): String(),
        t.Key("credential_id", optional=True): String(),
        t.Key("format", optional=True): String(),
    })

    _dataset_intake_settings = t.Dict({
        t.Key("dataset"): t.Type(Dataset),
        t.Key("dataset_version_id", optional=True): String(),
    })

    _dss_intake_settings = t.Dict({
        t.Key("project_id"): String(),
        t.Key("dataset_id", optional=True): String(),
        t.Key("partition", optional=True): t.Enum(
            TrainingDataSubsets.HOLDOUT,
            TrainingDataSubsets.VALIDATION,
            TrainingDataSubsets.ALL_BACKTESTS,
        ),
    })

    _jdbc_intake_settings = t.Dict({
        t.Key("data_store_id"): String(),
        t.Key("query", optional=True): String(),
        t.Key("table", optional=True): String(),
        t.Key("schema", optional=True): String(),
        t.Key("catalog", optional=True): String(),
        t.Key("fetch_size", optional=True): Int(),
        t.Key("credential_id", optional=True): String(),
    })

    _jdbc_output_settings = t.Dict({
        t.Key("data_store_id"): String(),
        t.Key("table"): String(),
        t.Key("schema", optional=True): String(),
        t.Key("catalog", optional=True): String(),
        t.Key("statement_type"): t.Enum(
            AVAILABLE_STATEMENT_TYPES.INSERT,
            AVAILABLE_STATEMENT_TYPES.UPDATE,
            AVAILABLE_STATEMENT_TYPES.INSERT_UPDATE,
            AVAILABLE_STATEMENT_TYPES.CREATE_TABLE,
        ),
        t.Key("update_columns", optional=True): t.List(String),
        t.Key("where_columns", optional=True): t.List(String),
        t.Key("credential_id", optional=True): String(),
        t.Key("create_table_if_not_exists", optional=True): t.Bool(),
    })

    _csv_settings = t.Dict({
        t.Key("delimiter", optional=True): t.Atom("tab") | String(min_length=1, max_length=1),
        t.Key("quotechar", optional=True): String(),
        t.Key("encoding", optional=True): String(),
    })

    _datasphere_settings = t.Dict({
        t.Key("data_store_id"): String(),
        t.Key("table", optional=True): String(),
        t.Key("schema", optional=True): String(),
        t.Key("credential_id", optional=True): String(),
    })

    @classmethod
    def validate_intake_settings(cls, input_intake_settings: Optional[IntakeSettings] = None) -> IntakeSettings:
        """Validates intake settings based on type run specific trafaret check
        Creates a copy in order to avoid mutating input data

        :param input_intake_settings:
        :return: A validated copy of intake settings
        """
        if input_intake_settings is None:
            intake_settings = IntakeSettings({"type": IntakeAdapters.LOCAL_FILE})
        else:
            # avoid mutating the input argument
            intake_settings = input_intake_settings.copy()

        # Validate the intake settings
        if intake_settings.get("type") not in IntakeAdapters:
            raise ValueError("Unsupported type parameter for intake_settings: {}".format(intake_settings.get("type")))

        elif intake_settings["type"] == IntakeAdapters.LOCAL_FILE:
            # This intake option requires us to upload the source
            # data ourselves

            if intake_settings.get("file") is None:
                raise ValueError(
                    "Missing source data. Either supply the `file` "
                    "parameter or switch to an intake option that does not "
                    "require it."
                )

        elif intake_settings["type"] == IntakeAdapters.S3:
            del intake_settings["type"]
            intake_settings = cls._s3_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.S3

        elif intake_settings["type"] == IntakeAdapters.GCP:
            del intake_settings["type"]
            intake_settings = cls._gcp_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.GCP

        elif intake_settings["type"] == IntakeAdapters.AZURE:
            del intake_settings["type"]
            intake_settings = cls._azure_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.AZURE

        elif intake_settings["type"] == IntakeAdapters.JDBC:
            del intake_settings["type"]
            intake_settings = cls._jdbc_intake_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.JDBC

        elif intake_settings["type"] == IntakeAdapters.DATASET:
            del intake_settings["type"]
            intake_settings = cls._dataset_intake_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.DATASET

            dataset = intake_settings["dataset"]
            intake_settings["dataset_id"] = dataset.id

            if "dataset_version_id" not in intake_settings:
                intake_settings["dataset_version_id"] = dataset.version_id

            del intake_settings["dataset"]

        elif intake_settings["type"] == IntakeAdapters.DSS:
            del intake_settings["type"]
            intake_settings = cls._dss_intake_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.DSS

        elif intake_settings["type"] == IntakeAdapters.DATASPHERE:
            del intake_settings["type"]
            intake_settings = cls._datasphere_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.DATASPHERE

        return intake_settings

    @classmethod
    def validate_output_settings(cls, input_output_settings: Optional[OutputSettings] = None) -> OutputSettings:
        """Validates output settings based on type run specific trafaret check
        Creates a copy in order to avoid mutating input data

        :param input_output_settings:
        :return: A validated copy of output settings
        """
        if input_output_settings is None:
            output_settings = OutputSettings({"type": OutputAdapters.LOCAL_FILE})
        else:
            output_settings = input_output_settings.copy()

        # Validate the output settings

        if output_settings.get("type") not in OutputAdapters:
            raise ValueError("Unsupported type parameter for output_settings: {}".format(output_settings.get("type")))
        elif output_settings["type"] == OutputAdapters.LOCAL_FILE:
            output_settings["path"] = output_settings.get("path")

        elif output_settings["type"] == OutputAdapters.S3:
            del output_settings["type"]
            output_settings = cls._s3_settings.check(output_settings)
            output_settings["type"] = OutputAdapters.S3

        elif output_settings["type"] == OutputAdapters.GCP:
            del output_settings["type"]
            output_settings = cls._gcp_settings.check(output_settings)
            output_settings["type"] = OutputAdapters.GCP

        elif output_settings["type"] == OutputAdapters.AZURE:
            del output_settings["type"]
            output_settings = cls._azure_settings.check(output_settings)
            output_settings["type"] = OutputAdapters.AZURE

        elif output_settings["type"] == OutputAdapters.JDBC:
            del output_settings["type"]
            output_settings = cls._jdbc_output_settings.check(output_settings)
            output_settings["type"] = OutputAdapters.JDBC

        elif output_settings["type"] == OutputAdapters.DATASPHERE:
            del output_settings["type"]
            output_settings = cls._datasphere_settings.check(output_settings)
            output_settings["type"] = OutputAdapters.DATASPHERE

        return output_settings

    @classmethod
    def _upload_intake_file(
        cls,
        intake_file: Dict[str, Any],
        intake_settings: IntakeSettings,
        job_response: Dict[str, Any],
        upload_read_timeout: int,
        output_file: Optional[typing.IO[Any]] = None,
    ) -> Optional[threading.Thread]:
        """Uploads intake file to running job, when intake type is localFile"""

        if intake_settings["type"] != IntakeAdapters.LOCAL_FILE:
            return None

        # There is source data to upload, so spin up a thread to handle
        # the upload concurrently and for thread safety issues, make
        # a copy of the REST client object
        _upload_client = cls._client.copy()
        job_csv_settings = job_response.get("jobSpec", {}).get("csvSettings", {})
        upload_thread = None

        def _get_file_size(fileobj: typing.BinaryIO) -> int:
            # To cover both files and filelike obj utilize .tell
            cur = fileobj.tell()
            fileobj.seek(0, os.SEEK_END)
            file_size = fileobj.tell()
            fileobj.seek(cur)

            return file_size

        # pylint: disable-next=unused-argument
        def _create_csv_chunk(
            header: Iterable[Any],
            reader: Iterator[Any],
            max_size: int,
            delimiter: str,
            encoding: str,
            quotechar: str,
        ) -> Tuple[io.StringIO, int]:
            chunk = io.StringIO()
            bytes_written = 0
            writer = csv.writer(chunk, delimiter=delimiter, quotechar=quotechar)
            writer.writerow(header)
            while bytes_written < max_size:
                try:
                    csv_chunk_content = next(reader)
                    written = writer.writerow(csv_chunk_content)
                    bytes_written += written
                except StopIteration:
                    break

            return chunk, bytes_written

        def _fileobj_to_csv_stream(fileobj: typing.BinaryIO, encoding: str) -> Iterator[Any]:
            stream = io.TextIOWrapper(fileobj, encoding=encoding)

            yield from csv.reader(stream)

            stream.close()

        def _upload_multipart(fileobj: typing.BinaryIO, base_upload_url: str) -> None:
            is_async = intake_settings.get("async", True)
            MAX_RETRY = 1 if is_async else 3
            MB_PER_CHUNK = 5
            CHUNK_MAX_SIZE = MB_PER_CHUNK * 1024 * 1024

            delimiter = job_csv_settings.get("delimiter", ",")
            encoding = job_csv_settings.get("encoding", "utf-8")
            quotechar = job_csv_settings.get("quotechar", '"')

            file_size = _get_file_size(fileobj)
            csv_stream = _fileobj_to_csv_stream(fileobj, encoding)

            # grab the header so it can be added on all parts
            header = next(csv_stream)

            part_number = 0
            bytes_written = 0
            while bytes_written <= file_size:
                part_upload_url = f"{base_upload_url}part/{part_number}/"

                # Read the inputfile in chunks of CHUNK_MAX_SIZE
                # Then call put multiple times increasing the part_number each time
                chunk, chunk_bytes = _create_csv_chunk(
                    header, csv_stream, CHUNK_MAX_SIZE, delimiter, encoding, quotechar
                )
                bytes_written += chunk_bytes
                if chunk_bytes == 0:
                    break

                for attempts in range(MAX_RETRY):
                    try:
                        chunk.seek(0)
                        response = _upload_client.put(
                            url=part_upload_url,
                            data=chunk,
                            headers={"content-type": "text/csv"},
                            timeout=(_upload_client.connect_timeout, upload_read_timeout),
                        )

                        # Success! don't retry
                        if response.status_code == 202:
                            chunk.close()
                            break
                    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
                        attempts += 1
                        if attempts == MAX_RETRY:
                            raise

                part_number += 1

            # finalize the upload to indicate no more data is arriving
            _upload_client.post(url=f"{base_upload_url}finalizeMultipart/")

        def _uploader() -> None:
            upload_url = job_response["links"]["csvUpload"]

            if "file_path" in intake_file:
                fileobj = open(  # pylint: disable=consider-using-with
                    intake_file["file_path"], "rb"
                )
            else:
                fileobj = intake_file["filelike"]
                fileobj.seek(0)

            try:
                if intake_settings.get("multipart"):
                    _upload_multipart(fileobj, upload_url)

                else:
                    _upload_client.put(
                        url=upload_url,
                        data=fileobj,
                        headers={"content-type": "text/csv"},
                        timeout=(_upload_client.connect_timeout, upload_read_timeout),
                    )
            finally:
                if hasattr(fileobj, "close"):
                    fileobj.close()

        if output_file is not None:
            # If output_file is specified, we upload and download
            # concurrently

            upload_thread = threading.Thread(target=_uploader)
            upload_thread.daemon = True
            upload_thread.start()

        else:
            # Otherwise, upload is synchronous

            _uploader()
        return upload_thread

    @classmethod
    def _job_path(cls, project_id: str, job_id: str) -> str:
        raise NotImplementedError

    def _delete(self, ignore_404_errors: bool = False) -> None:
        """
        Cancel this job. If this job has not finished running, it will be
        removed and canceled.
        """
        status = self._get_status()

        prediction_job_id = status["links"]["self"].split("/")[-2]
        try:
            self._client.delete(self._job_path(project_id="", job_id=prediction_job_id))
        except errors.ClientError as exc:
            if exc.status_code == 404 and ignore_404_errors:
                return
            raise

    def _get_status(self) -> Any:
        """Get status of batch job"""
        batch_job = super().get(project_id="", job_id=self.id)
        batch_job.id = self.id

        return batch_job._safe_data

    def _download(self, fileobj: typing.IO[Any], timeout: int = 120, read_timeout: int = 660) -> None:
        """Downloads the CSV result of a batch job"""

        status = self._wait_for_download(timeout=timeout)
        download_iter = self._client.get(
            status["links"]["download"],
            stream=True,
            timeout=read_timeout,
        ).iter_content(chunk_size=8192)

        for chunk in download_iter:
            if chunk:
                fileobj.write(chunk)

        # Check if job was aborted during download (and the download is incomplete)
        status = self._get_status()
        if status["status"] in ("ABORTED", "FAILED"):
            raise RuntimeError("Job {} was aborted: {}".format(self.id, status["status_details"]))

    def _wait_for_download(self, timeout: int = 120) -> Any:
        """Waits for download to become available"""
        start = time.time()
        status = None
        while True:
            status = self._get_status()

            output_adapter_type = status["job_spec"].get("output_settings", {}).get("type")
            if output_adapter_type and not output_adapter_type == OutputAdapters.LOCAL_FILE:
                raise RuntimeError(
                    (
                        "You cannot download predictions from jobs that did not use local_file as "
                        "the output adapter. Job with ID {} had the output adapter defined as {}."
                    ).format(self.id, output_adapter_type)
                )

            if status["status"] in ("ABORTED", "FAILED"):
                raise RuntimeError("Job {} was aborted: {}".format(self.id, status["status_details"]))

            if "download" in status["links"]:
                break

            if timeout >= 0 and time.time() - start > timeout:  # pylint: disable=chained-comparison
                break

            time.sleep(1)

        if "download" not in status["links"]:
            # Ignore 404 errors here if the job never started - then we can't abort it
            self._delete(ignore_404_errors=True)
            raise RuntimeError(
                (
                    "Timed out waiting for download to become available for job ID {}. "
                    "Other jobs may be occupying the queue. Consider raising the timeout."
                ).format(self.id)
            )

        return status



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc. Confidential.
#
# This is unpublished proprietary source code of DataRobot, Inc.
# and its affiliates.
#
# The copyright notice above does not evidence any actual or intended
# publication of such source code.
from __future__ import annotations

from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot.models.api_object import APIObject


class OtelMetricSummary(APIObject):
    """Reported OpenTelemetry metric summary information.

    .. versionadded:: v3.9

    Attributes
    ----------
    otelName: str
        Name of the reported metric.
    description: Optional[str]
        Description of the reported metric.
    metricType: Optional[str]
        Reported metric type (e.g. counter, gauge, histogram).
    units: Optional[str]
        Units of the reported metric.
    """

    _path = "otel/{}/{}/metrics/summary/"
    _converter = t.Dict({
        t.Key("otel_name"): t.String(),
        t.Key("description", optional=True): t.Or(t.String(), t.Null()),
        t.Key("metric_type", optional=True): t.Or(t.String(), t.Null()),
        t.Key("units", optional=True): t.Or(t.String(), t.Null()),
    }).ignore_extra("*")

    def __init__(
        self,
        otel_name: str,
        description: Optional[str] = None,
        metric_type: Optional[str] = None,
        units: Optional[str] = None,
    ) -> None:
        self.otel_name = otel_name
        self.description = description
        self.metric_type = metric_type
        self.units = units

    def __repr__(self) -> str:
        params = [self.otel_name]
        if self.units:
            params.append(f"units={self.units}")
        if self.metric_type:
            params.append(f"type={self.metric_type}")
        if self.description:
            params.append(f"description='{self.description}'")
        return f"OtelMetricSummary({','.join(params)})"

    @classmethod
    def list(
        cls,
        entity_type: str,
        entity_id: str,
        search: Optional[str] = None,
        metric_type: Optional[str] = None,
    ) -> List[OtelMetricSummary]:
        """List OpenTelemetry metric summary information.

        .. versionadded:: v3.9

        Parameters
        ----------
        entity_type: str
            The entity type of the reported metrics (e.g. deployment, or use_case).
        entity_id: str
            The entity ID of the reported metrics (e.g. `123456`).
        search: Optional[str]
            Only return metrics whose name contains this case-sensitive value.
        metric_type: Optional[str]
            Only return metrics whose type matches this value (e.g. counter, gauge, histogram).

        Returns
        -------
        summary: List[OtelMetricSummary]
        """
        path = cls._path.format(entity_type, entity_id)
        params: Dict[str, Any] = {}
        if search:
            params["search"] = search
        if metric_type:
            params["metricType"] = metric_type

        data = cls._client.get(path, params=params if params else None).json()["data"]
        return [cls.from_server_data(d) for d in data]



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate


class ResourceBundle(APIObject):
    """A DataRobot resource bundle.

    A resource bundle provides a commonly available id that allows users to specify the
    required resources to be used. This includes CPU/GPU and memory resources.

    .. versionadded:: v3.8

    Attributes
    ----------
    id: str
        The ID of the resource bundle.
    name: str
         A short name for the bundle.
    description: Optional[str]
        A short description of CPU, Memory and other resources.
    cpu_count: float
        Maximum number of CPUs available.
    memory_bytes: int
        Maximum amount of memory available.
    gpu_maker: Optional[str]
        The manufacturer of the GPU (e.g. nvidia, amd, intel)
    gpu_count: Optional[float]
        Maximum number of GPUs available.
    gpu_memory_bytes: int
        Maximum amount of GPU memory available.
    use_cases: List[str]
        List of use cases this bundle supports (e.g. customApplication, customJob, customModel, sapAICore).
    is_default: Optional[bool]
        If this should be the default resource choice.
    is_deleted: Optional[bool]
        If the bundle has been deleted and should not be used.
    has_gpu: bool
        If this bundle provides at least one GPU resource.
    """

    _path = "mlops/compute/bundles/"
    _converter = t.Dict({
        t.Key("id"): t.String(),
        t.Key("name"): t.String(),
        t.Key("description"): t.Or(t.String(), t.Null()),
        t.Key("cpu_count"): t.Float(),
        t.Key("memory_bytes"): t.Int(),
        t.Key("gpu_maker", optional=True): t.Or(t.String(), t.Null()),
        t.Key("gpu_count", optional=True): t.Or(t.Float(), t.Null()),
        t.Key("gpu_memory_bytes"): t.Int(),
        t.Key("use_cases"): t.List(t.String()),
        t.Key("is_default", optional=True): t.Bool(),
        t.Key("is_deleted", optional=True): t.Bool(),
        t.Key("has_gpu"): t.Bool(),
    }).ignore_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        cpu_count: float,
        memory_bytes: int,
        gpu_memory_bytes: int,
        use_cases: List[str],
        has_gpu: bool,
        description: Optional[str] = None,
        gpu_maker: Optional[int] = None,
        gpu_count: Optional[float] = None,
        is_default: Optional[bool] = None,
        is_deleted: Optional[bool] = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.cpu_count = cpu_count
        self.memory_bytes = memory_bytes
        self.gpu_maker = gpu_maker
        self.gpu_count = gpu_count
        self.gpu_memory_bytes = gpu_memory_bytes
        self.use_cases = use_cases
        self.is_default = is_default
        self.is_deleted = is_deleted
        self.has_gpu = has_gpu

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name}, id={self.id})"

    @classmethod
    def list(
        cls,
        use_cases: Optional[List[str]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List["ResourceBundle"]:
        """
        List all resource bundles.

        .. versionadded:: v3.8

        Parameters
        ----------
        use_cases : List[str], optional
            List of use-cases
        offset: Optional[int]
            Offset for pagination.
        limit: Optional[int]
            Limit for pagination.

        Returns
        -------
        resource_bundles: List[ResourceBundle]
        """
        params: Dict[str, Any] = {}
        if use_cases:
            params["use_cases"] = use_cases
        if offset is not None:
            params["offset"] = offset
        if limit is not None:
            params["limit"] = limit

        data = unpaginate(cls._path, params, cls._client)
        return [cls.from_server_data(d) for d in data]

    @classmethod
    def get(cls, bundle_id: str) -> "ResourceBundle":
        """
        Get the specified resource bundle.

        .. versionadded:: v3.8

        Parameters
        ----------
        bundle_id: str
            ID of the resource bundle (e.g. cpu.micro)
        """
        response = cls._client.get(f"{cls._path}{bundle_id}/")
        return cls.from_server_data(response.json())



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc. Confidential.
#
# This is unpublished proprietary source code of DataRobot, Inc.
# and its affiliates.
#
# The copyright notice above does not evidence any actual or intended
# publication of such source code.
from __future__ import annotations

from enum import Enum


class MetricAggregation(str, Enum):
    """Metric aggregation options."""

    AVERAGE = "average"
    MAX = "max"
    MIN = "min"
    SUM = "sum"


class MetricResolution(str, Enum):
    ONE_MINUTE = "PT1M"
    FIVE_MINUTES = "PT5M"
    ONE_HOUR = "PT1H"
    ONE_DAY = "P1D"



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from .config import DataRobotAppFrameworkBaseSettings, getenv

__all__ = ["getenv", "DataRobotAppFrameworkBaseSettings"]



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Optional

import dateutil
import pandas as pd
import trafaret as t

from datarobot._compat import String
from datarobot.enums import ACCURACY_METRIC, BUCKET_SIZE
from datarobot.models.api_object import APIObject
from datarobot.models.deployment.mixins import MonitoringDataQueryBuilderMixin
from datarobot.utils import from_api
from datarobot.utils.deprecation import deprecated

if TYPE_CHECKING:
    from datarobot._compat import TypedDict

    class Period(TypedDict, total=False):
        start: datetime
        end: datetime

    class Metric(TypedDict):
        model_id: str
        metric_name: str
        value: int
        sample_size: int
        baseline_value: int
        baseline_sample_size: int
        percent_change: int

    class Summary(TypedDict, total=False):
        period: Period

    class Bucket(TypedDict):
        period: Period
        value: int
        sample_size: int

    class ClassDistribution(TypedDict):
        class_name: str
        count: int
        percent: float

    class PredictionsVsActualsSummaryBucket(TypedDict):
        row_count_total: int
        row_count_with_actual: int

    class GeoPoint(TypedDict):
        latitude: float
        longitude: float

    class PredictionsVsActualsBaselineBucket(TypedDict):
        model_id: str
        row_count_total: int
        row_count_with_actual: int
        mean_predicted_value: Optional[float]
        mean_actual_value: Optional[float]
        predicted_class_distribution: Optional[List[ClassDistribution]]
        actual_class_distribution: Optional[List[ClassDistribution]]
        mean_predicted_geo_centroid: Optional[GeoPoint]
        mean_actual_geo_centroid: Optional[GeoPoint]

    class PredictionsVsActualsOverTimeBucket(TypedDict):  # pylint: disable=missing-class-docstring
        period: Period
        model_id: str
        row_count_total: int
        row_count_with_actual: int
        mean_predicted_value: Optional[float]
        mean_actual_value: Optional[float]
        predicted_class_distribution: Optional[List[ClassDistribution]]
        actual_class_distribution: Optional[List[ClassDistribution]]
        mean_predicted_geo_centroid: Optional[GeoPoint]
        mean_actual_geo_centroid: Optional[GeoPoint]


class Accuracy(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment accuracy information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve accuracy metrics
    period : dict
        the time period used to retrieve accuracy metrics
    metrics : dict
        the accuracy metrics
    """

    _path = "deployments/{}/accuracy/"
    _converter = t.Dict({
        t.Key("period"): t.Dict({
            t.Key("start"): String >> dateutil.parser.parse,
            t.Key("end"): String >> dateutil.parser.parse,
        }),
        t.Key("data"): t.List(t.Dict().allow_extra("*")),
        t.Key("model_id", optional=True): t.Or(t.String(), t.Null),
        t.Key("metrics", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
    }).allow_extra("*")

    def __init__(
        self,
        period: Optional[Period] = None,
        data: Optional[List[Metric]] = None,
        metrics: Optional[Dict[str, Metric]] = None,
        model_id: Optional[str] = None,
    ) -> None:
        self.period = period
        self.data = data
        self._metrics = metrics
        self._model_id = model_id

    def __repr__(self) -> str:
        if self.period:
            start = self.period.get("start")
            end = self.period.get("end")
        else:
            start, end = "Unknown", "Unknown"  # type: ignore [assignment]
        return f"{self.__class__.__name__}({start} - {end})"

    def __getitem__(self, item: str) -> Optional[int]:
        if self.metrics and item in self.metrics.keys():
            return self.metrics[item].get("value")
        return None

    @classmethod
    def get(
        cls,
        deployment_id: str,
        model_id: Optional[str | List[str]] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        target_classes: Optional[List[str]] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
        metric: Optional[str] = None,
        baseline_model_id: Optional[str] = None,
    ) -> Accuracy:
        """Retrieve values of accuracy metrics over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : Optional[str | List[str]]
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        target_classes : list[str], optional
            Optional list of target class strings
        segment_attribute : Optional[str]
            (New in Version v3.6) the segment attribute
        segment_value : Optional[str]
            (New in Version v3.6) the segment value
        metric : str
            (New in Version v3.9) the metric value to retrieve,
            must be provided when querying for multiple models
        baseline_model_id : str
            (New in Version v3.9) the id of the baseline model when calculating percentage change

        Returns
        -------
        accuracy : Accuracy
            the queried accuracy metrics information

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment, Accuracy
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            accuracy = Accuracy.get(deployment.id)
            accuracy.period['end']
            >>>'2019-08-01 00:00:00+00:00'
            accuracy.metric['LogLoss']['value']
            >>>0.7533
            accuracy.metric_values['LogLoss']
            >>>0.7533
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            target_class=target_classes,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
            metric=metric,
            baseline_model_id=baseline_model_id,
        )
        data = cls._client.get(path, params=params).json()

        # we don't want to convert keys of the metrics object
        metrics = data.pop("metrics", {})
        for metric_name, value in metrics.items():
            metrics[metric_name] = from_api(value, keep_null_keys=True)

        data = from_api(data, keep_null_keys=True)
        data["metrics"] = metrics  # type: ignore[call-overload]
        return cls.from_data(data)

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`metrics` is deprecated, use `data` instead.",
    )
    def metrics(self) -> Optional[Dict[str, Metric]]:
        return self._metrics

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`model_id` is deprecated, inspect `model_id` inside `data` instead.",
    )
    def model_id(self) -> Optional[str]:
        return self._model_id

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`metric_values` is deprecated, inspect `value` inside `data` instead.",
    )
    def metric_values(self) -> Dict[str, Optional[int]]:
        """The value for all metrics, keyed by metric name.

        Returns
        -------
        metric_values: Dict
        """

        return {name: value.get("value") for name, value in self.metrics.items()}

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`metric_baselines` is deprecated, inspect `baseline_value` inside `data` instead.",
    )
    def metric_baselines(self) -> Dict[str, Optional[int]]:
        """The baseline value for all metrics, keyed by metric name.

        Returns
        -------
        metric_baselines: Dict
        """

        return {name: value.get("baseline_value") for name, value in self.metrics.items()}

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`percent_changes` is deprecated, inspect `percent_change` inside `data` instead.",
    )
    def percent_changes(self) -> Dict[str, Optional[int]]:
        """The percent change of value over baseline for all metrics, keyed by metric name.

        Returns
        -------
        percent_changes: Dict
        """

        return {name: value.get("percent_change") for name, value in self.metrics.items()}


class AccuracyOverTime(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment accuracy over time information.

    Attributes
    ----------
    metric : str
        the accuracy metric being retrieved
    buckets : list
        how the accuracy metric changes over time
    baselines : list
        baselines for the accuracy metric
    """

    _path = "deployments/{}/accuracyOverTime/"
    _period = t.Dict({
        t.Key("start"): String >> dateutil.parser.parse,
        t.Key("end"): String >> dateutil.parser.parse,
    })
    _bucket = t.Dict({t.Key("period"): t.Or(_period, t.Null)}).allow_extra("*")
    _converter = t.Dict({
        t.Key("buckets"): t.List(_bucket),
        t.Key("summary", optional=True): t.Or(_bucket, t.Null),
        t.Key("baseline", optional=True): t.Or(_bucket, t.Null),
        t.Key("baselines"): t.List(t.Dict().allow_extra("*")),
        t.Key("metric"): String(),
        t.Key("model_id", optional=True): t.Or(String(), t.Null),
        t.Key("segment_attribute", optional=True): t.String(),
        t.Key("segment_value", optional=True): t.Or(String(allow_blank=True), t.Null),
    }).allow_extra("*")

    def __init__(
        self,
        buckets: Optional[List[Bucket]] = None,
        summary: Optional[Summary] = None,
        baseline: Optional[Bucket] = None,
        baselines: Optional[Bucket] = None,
        metric: Optional[str] = None,
        model_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ):
        self.buckets = buckets if buckets is not None else []
        self._summary = summary
        self._baseline = baseline
        self.baselines = baselines if baselines is not None else []
        self.metric = metric
        self._model_id = model_id
        self.segment_attribute = segment_attribute
        self.segment_value = segment_value

    def __repr__(self) -> str:
        if self.buckets:
            start = self.buckets[0]["period"]["start"]
            end = self.buckets[-1]["period"]["end"]
        else:
            start, end = "Unknown", "Unknown"  # type: ignore [assignment]
        return f"{self.__class__.__name__}({self.metric} | {start} - {end})"

    @classmethod
    def get(
        cls,
        deployment_id: str,
        metric: Optional[ACCURACY_METRIC] = None,
        model_id: Optional[str | List[str]] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        bucket_size: Optional[str] = None,
        target_classes: Optional[List[str]] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> AccuracyOverTime:
        """Retrieve information about how an accuracy metric changes over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        metric : ACCURACY_METRIC
            the accuracy metric to retrieve
        model_id : Optional[str | List[str]]
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        bucket_size : str
            time duration of a bucket, in ISO 8601 time duration format
        target_classes : list[str], optional
            Optional list of target class strings
        segment_attribute : Optional[str]
            (New in Version v3.6) the segment attribute
        segment_value : Optional[str]
            (New in Version v3.6) the segment value

        Returns
        -------
        accuracy_over_time : AccuracyOverTime
            the queried accuracy metric over time information

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment, AccuracyOverTime
            from datarobot.enums import ACCURACY_METRICS
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            accuracy_over_time = AccuracyOverTime.get(deployment.id, metric=ACCURACY_METRIC.LOGLOSS)
            accuracy_over_time.metric
            >>>'LogLoss'
            accuracy_over_time.metric_values
            >>>{datetime.datetime(2019, 8, 1): 0.73, datetime.datetime(2019, 8, 2): 0.55}
        """
        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            metric=metric,
            bucket_size=bucket_size,
            target_class=target_classes,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        return cls.from_data(case_converted)

    @classmethod
    def get_as_dataframe(
        cls,
        deployment_id: str,
        metrics: Optional[List[Optional[ACCURACY_METRIC]]] = None,
        model_id: Optional[str | List[str]] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        bucket_size: Optional[str] = None,
    ) -> pd.DataFrame:
        """Retrieve information about how a list of accuracy metrics change over
        a certain time period as pandas DataFrame.

        In the returned DataFrame, the columns corresponds to the metrics being retrieved;
        the rows are labeled with the start time of each bucket.

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        metrics : [ACCURACY_METRIC]
            the accuracy metrics to retrieve
        model_id : Optional[str | List[str]]
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        bucket_size : str
            time duration of a bucket, in ISO 8601 time duration format

        Returns
        -------
        accuracy_over_time: pd.DataFrame
        """
        if not metrics:
            if metrics == []:
                raise ValueError("Metrics must be a list of ACCURACY_METRIC or None, but cannot be an empty list")
            metrics = [None]
        series = {}
        for metric in metrics:
            fetched = AccuracyOverTime.get(
                deployment_id,
                model_id=model_id,
                metric=metric,
                start_time=start_time,
                end_time=end_time,
                bucket_size=bucket_size,
            )
            if not fetched.buckets:
                continue
            dataframe = pd.json_normalize(fetched.buckets)
            dataframe.set_index(
                ["model_id", "period.start"],
                drop=True,
                inplace=True,  # noqa: PD002
            )
            series[fetched.metric] = dataframe["value"]
        if series:
            return pd.DataFrame.from_dict(series)
        else:
            return pd.DataFrame()

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`summary` is deprecated, use `Deployment.get_accuracy` instead.",
    )
    def summary(self) -> Optional[Summary]:
        return self._summary

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`baseline` is deprecated, inspect `baselines` instead.",
    )
    def baseline(self) -> Optional[Bucket]:
        return self._baseline

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`model_id` is deprecated, inspect `model_id` inside `buckets` instead.",
    )
    def model_id(self) -> Optional[str]:
        return self._model_id

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`bucket_values` is deprecated, inspect `value` inside `buckets` instead.",
    )
    def bucket_values(self) -> Dict[datetime, int]:
        """The metric value for all time buckets, keyed by start time of the bucket.

        Returns
        -------
        bucket_values: Dict
        """
        if self.buckets:
            return {bucket["period"]["start"]: bucket["value"] for bucket in self.buckets if bucket.get("period")}
        return {}

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`bucket_values` is deprecated, inspect `sample_size` inside `buckets` instead.",
    )
    def bucket_sample_sizes(self) -> Dict[datetime, int]:
        """The sample size for all time buckets, keyed by start time of the bucket.

        Returns
        -------
        bucket_sample_sizes: Dict
        """
        if self.buckets:
            return {bucket["period"]["start"]: bucket["sample_size"] for bucket in self.buckets if bucket.get("period")}
        return {}


class PredictionsVsActualsOverTime(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment predictions vs actuals over time information.

    Attributes
    ----------
    summary : dict
        predictions vs actuals over time summary for all models and buckets queried
    baselines : List
        target baseline for each model queried
    buckets : List
        predictions vs actuals over time bucket for each model and bucket queried
    segment_attribute : Optional[str]
        (New in Version v3.6) the segment attribute
    segment_value : Optional[str]
        (New in Version v3.6) the segment value

    """

    _path = "deployments/{}/predictionsVsActualsOverTime/"
    _period = t.Dict({
        t.Key("start"): String >> dateutil.parser.parse,
        t.Key("end"): String >> dateutil.parser.parse,
    })
    _converter = t.Dict({
        t.Key("summary"): t.Dict().allow_extra("*"),
        t.Key("baselines"): t.List(t.Dict().allow_extra("*")),
        t.Key("buckets"): t.List(t.Dict({"period": _period}).allow_extra("*")),
        t.Key("segment_attribute", optional=True): t.String(),
        t.Key("segment_value", optional=True): t.Or(String(allow_blank=True), t.Null),
    })

    def __init__(
        self,
        summary: Optional[PredictionsVsActualsSummaryBucket] = None,
        baselines: Optional[List[PredictionsVsActualsBaselineBucket]] = None,
        buckets: Optional[List[PredictionsVsActualsOverTimeBucket]] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ):
        self.summary = summary
        self.baselines = baselines
        self.buckets = buckets
        self.segment_attribute = segment_attribute
        self.segment_value = segment_value

    @classmethod
    def get(
        cls,
        deployment_id: str,
        model_ids: Optional[List[str]] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        bucket_size: Optional[BUCKET_SIZE] = None,
        target_classes: Optional[List[str]] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> PredictionsVsActualsOverTime:
        """Retrieve information for deployment's predictions vs actuals over a certain time period.

        .. versionadded:: v3.3

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_ids : list[str]
            ID of models to retrieve predictions vs actuals stats
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        bucket_size : BUCKET_SIZE
            time duration of each bucket
        target_classes : list[str]
            class names of target, only for deployments with multiclass target
        segment_attribute : Optional[str]
            (New in Version v3.6) the segment attribute
        segment_value : Optional[str]
            (New in Version v3.6) the segment value


        Returns
        -------
        predictions_vs_actuals_over_time : PredictionsVsActualsOverTime
            the queried predictions vs actuals over time information
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_ids,
            bucket_size=bucket_size,
            target_class=target_classes,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )

        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        return cls.from_data(case_converted)



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# limitations under the License.
from __future__ import annotations

import asyncio
import functools
from typing import Any, Callable, Coroutine, TypeVar

from typing_extensions import ParamSpec

T_Retval = TypeVar("T_Retval")
T_ParamSpec = ParamSpec("T_ParamSpec")


def syncify(
    async_func: Callable[T_ParamSpec, Coroutine[Any, Any, T_Retval]],
) -> "Callable[T_ParamSpec, T_Retval]":
    """
    Decorator to run an async function in a synchronous context.
    """

    @functools.wraps(async_func)
    def sync_wrapper(*args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs) -> T_Retval:
        try:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                raise RuntimeError("Cannot call a synchronous method from a running event loop.")
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        return loop.run_until_complete(async_func(*args, **kwargs))

    return sync_wrapper



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc. Confidential.
#
# This is unpublished proprietary source code of DataRobot, Inc.
# and its affiliates.
#
# The copyright notice above does not evidence any actual or intended
# publication of such source code.
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional

import dateutil
import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.otel.metric_enums import MetricResolution

OtelMetricValueTrafaret = t.Dict({
    t.Key("otel_name"): t.String(),
    t.Key("samples"): t.Int(),
    t.Key("delta", optional=True): t.Or(t.Float(), t.Null()),
    t.Key("value", optional=True): t.Or(t.Float(), t.Null()),
    t.Key("buckets", optional=True): t.Or(t.List(t.Dict().allow_extra("*")), t.Null()),
    t.Key("display_name", optional=True): t.Or(t.String(), t.Null()),
    t.Key("aggregation", optional=True): t.Or(t.String(), t.Null()),
    t.Key("id", optional=True): t.Or(t.String(), t.Null()),
}).ignore_extra("*")


class OtelMetricBucketValue(APIObject):
    """OpenTelemetry metric bucket value.

    .. versionadded:: v3.9

    Attributes
    ----------
    otel_name: str
        The OTEL name of the metric.
    value: Optional[float]
        The metric value for the period.
    buckets: Optional[list[dict]
        Histogram bucket values.
    samples: int
        Number of metric values in the period.
    delta: Optional[float]
        Difference between value from previous period (if any).
    display_name: Optional[str]
        The display name of the metric.
    aggregation: Optional[str]
        The aggregation type of the metric.
    id: Optional[str]
        The ID of the metric configuration (if any).
    """

    _converter = OtelMetricValueTrafaret

    def __init__(
        self,
        otel_name: str,
        samples: int,
        value: Optional[float] = None,
        buckets: Optional[List[Dict[str, Any]]] = None,
        delta: Optional[float] = None,
        display_name: Optional[str] = None,
        aggregation: Optional[str] = None,
        id: Optional[str] = None,
    ) -> None:
        self.otel_name = otel_name
        self.value = value
        self.buckets = buckets
        self.display_name = display_name
        self.samples = samples
        self.delta = delta
        self.aggregation = aggregation
        self.id = id

    def __repr__(self) -> str:
        params = [
            self.display_name or self.otel_name,
            str(self.value),
        ]
        return f"{self.__class__.__name__}({', '.join(params)})"


class OtelMetricValue(APIObject):
    """OpenTelemetry metric value.

    .. versionadded:: v3.9

    Attributes
    ----------
    start_time: datetime
        Start time of the metric value period.
    end_time: datetime
        End time of the metric value period.
    values: List[OtelMetricBucketValue]
        List of metric bucket values.
    """

    _path = "otel/{}/{}/metrics/valuesOverTime/"
    _converter = t.Dict({
        t.Key("start_time"): t.String() >> dateutil.parser.parse,
        t.Key("end_time"): t.String() >> dateutil.parser.parse,
        t.Key("values"): t.List(OtelMetricValueTrafaret),
    }).ignore_extra("*")

    def __init__(
        self,
        start_time: datetime,
        end_time: datetime,
        values: List[OtelMetricBucketValue | dict[str, Any]],
    ) -> None:
        self.start_time = start_time
        self.end_time = end_time
        self.values = []
        for value in values:
            if isinstance(value, OtelMetricBucketValue):
                self.values.append(value)
            if isinstance(value, dict):
                self.values.append(OtelMetricBucketValue(**value))

    def __repr__(self) -> str:
        params = [
            f"start={self.start_time}",
            f"{len(self.values)} values",
        ]
        return f"{self.__class__.__name__}({', '.join(params)})"

    @classmethod
    def list(
        cls,
        entity_type: str,
        entity_id: str,
        resolution: MetricResolution,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
    ) -> List[OtelMetricValue]:
        """List OpenTelemetry metric buckets with values.

        .. versionadded:: v3.9

        Parameters
        ----------
        entity_type: str
            The entity type of the reported metrics (e.g. deployment, or use_case).
        entity_id: str
            The entity ID of the reported metrics (e.g. `123456`).
        resolution: OtelMetricResolution
            Period for values of the metric list.
        start_time: Optional[datetime]
            Start time of the metric list.
        end_time: Optional[datetime]
            End time of the metric list.

        Returns
        -------
        info: List[OtelMetricValue]
        """
        path = cls._path.format(entity_type, entity_id)
        params: Dict[str, Any] = {
            "resolution": resolution,
        }
        if start_time:
            params["startTime"] = start_time.isoformat()
        if end_time:
            params["endTime"] = end_time.isoformat()

        data = cls._client.get(path, params=params).json()["data"]
        return [cls.from_server_data(d) for d in data]



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc. Confidential.
#
# This is unpublished proprietary source code of DataRobot, Inc.
# and its affiliates.
#
# The copyright notice above does not evidence any actual or intended
# publication of such source code.
from __future__ import annotations

from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.otel.metric_enums import MetricAggregation
from datarobot.utils import to_api
from datarobot.utils.pagination import unpaginate


class OtelMetricConfig(APIObject):
    """OpenTelemetry metric configuration.

    .. versionadded:: v3.9

    Attributes
    ----------
    otelName: str
        Name of the OpenTelemetry reported metric.
    id: Optional[str]
        ID of the metric configuration. This is optional to allow setting list.
    displayName: Optional[str]
        Name to display for the reported metric.
    aggregation: Optional[str|MetricAggregation]
        Aggregation type to use for the reported metric.
    enabled: Optional[bool]
        Whether the reported metric is shown in displays.
    """

    _path = "otel/{}/{}/metrics/configs/"
    _converter = t.Dict({
        t.Key("otel_name"): t.String(),
        t.Key("id", optional=True): t.Or(t.String(), t.Null()),
        t.Key("display_name", optional=True): t.Or(t.String(), t.Null()),
        t.Key("aggregation", optional=True): t.Or(t.Enum(*[e.value for e in MetricAggregation]), t.String(), t.Null()),
        t.Key("enabled", optional=True): t.Or(t.Bool(), t.Null()),
    }).ignore_extra("*")

    def __init__(
        self,
        otel_name: str,
        id: Optional[str] = None,
        display_name: Optional[str] = None,
        aggregation: Optional[str | MetricAggregation] = None,
        enabled: Optional[bool] = None,
        # NOTE: fields below are not provided by server, but tracked for object operations
        entity_type: Optional[str] = None,
        entity_id: Optional[str] = None,
    ) -> None:
        self.id = id
        self.otel_name = otel_name
        self.display_name = display_name
        self.aggregation = aggregation
        self.enabled = enabled
        self.entity_type = entity_type
        self.entity_id = entity_id

    def __repr__(self) -> str:
        name = self.display_name
        if not name:
            name = self.otel_name
            if self.aggregation:
                name += f" ({self.aggregation})"
        return f"OtelMetricConfig({name}, {self.id})"

    @classmethod
    def from_server_data_with_entity(
        cls,
        data: Dict[str, Any],
        entity_type: Optional[str] = None,
        entity_id: Optional[str] = None,
    ) -> OtelMetricConfig:
        config = cls.from_server_data(data)
        config.entity_type = entity_type
        config.entity_id = entity_id
        return config

    @classmethod
    def create(
        cls,
        entity_type: str,
        entity_id: str,
        otel_name: str,
        display_name: Optional[str] = None,
        aggregation: Optional[str] = None,
        enabled: Optional[bool] = None,
    ) -> OtelMetricConfig:
        """Create a new OpenTelemetry metric configuration.

        .. versionadded:: v3.9

        Parameters
        ----------
        entity_type: str
            The entity type of the log entries (e.g. deployment, or use_case)
        entity_id: str
            The entity id of the log entries (e.g. `123456`)
        otel_name: str
            Name of the reported metric.
        display_name: Optional[str]
            Name to display for the reported metric.
        aggregation: Optional[str]
            Aggregation type to use for the reported metric.
        enabled: Optional[bool]
            Whether the reported metric is shown in displays.
        """
        body: Dict[str, Any] = {
            "otelName": otel_name,
        }
        if display_name:
            body["displayName"] = display_name
        if aggregation:
            body["aggregation"] = str(aggregation)
        if enabled is not None:
            body["enabled"] = enabled
        url = cls._path.format(entity_type, entity_id)
        response = cls._client.post(url, json=body)
        return cls.from_server_data_with_entity(response.json(), entity_type, entity_id)

    @classmethod
    def set_list(
        cls,
        entity_type: str,
        entity_id: str,
        configs: List[OtelMetricConfig | Dict[str, Any]],
    ) -> List[OtelMetricConfig]:
        """Set a list of OpenTelemetry metric configurations.

        .. versionadded:: v3.9

        Parameters
        ----------
        entity_type: str
            The entity type of the log entries (e.g. deployment, or use_case)
        entity_id: str
            The entity id of the log entries (e.g. `123456`)
        configs: List[OtelMetricConfig]
            Ordered list of OpenTelemetry metric configurations.
        """
        items: List[Dict[str, Any]] = []
        for config in configs:
            item: Dict[str, Any] = {}
            if isinstance(config, dict):
                item = config
            elif isinstance(config, APIObject):
                data = to_api(config)
                if isinstance(data, dict):
                    item = data
                else:
                    continue
            item.pop("id", None)
            item.pop("entityId", None)
            item.pop("entityType", None)
            items.append(item)

        url = cls._path.format(entity_type, entity_id)
        cls._client.put(url, json={"values": items})

        return cls.list(entity_type, entity_id)

    @classmethod
    def list(
        cls,
        entity_type: str,
        entity_id: str,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[OtelMetricConfig]:
        """List OpenTelemetry metric configurations.

        .. versionadded:: v3.9

        Parameters
        ----------
        entity_type: str
            The entity type of the log entries (e.g. deployment, or use_case)
        entity_id: str
            The entity id of the log entries (e.g. `123456`)
        offset: Optional[int]
            Offset for pagination.
        limit: Optional[int]
            Limit for pagination.

        Returns
        -------
        info: List[OtelMetricConfig]
        """
        path = cls._path.format(entity_type, entity_id)
        params: Dict[str, Any] = {}
        if offset:
            params["offset"] = offset
        if limit:
            params["limit"] = limit

        if offset is None:
            data = unpaginate(path, params, cls._client)
        else:
            response = cls._client.get(path, params=params if params else None)
            data = response.json()["data"]

        return [cls.from_server_data_with_entity(d, entity_type, entity_id) for d in data]

    @classmethod
    def get(
        cls,
        entity_type: str,
        entity_id: str,
        config_id: str,
    ) -> OtelMetricConfig:
        """Get an OpenTelemetry metric configuration by ID.

        .. versionadded:: v3.9

        Parameters
        ----------
        entity_type: str
            The entity type of the log entries (e.g. deployment, or use_case)
        entity_id: str
            The entity id of the log entries (e.g. `123456`)
        config_id: str
            ID of the OpenTelemetry metric configuration.

        Returns
        -------
        config : OtelMetricConfig
        """
        url = cls._path.format(entity_type, entity_id) + config_id + "/"
        response = cls._client.get(url)
        return cls.from_server_data_with_entity(response.json(), entity_type, entity_id)

    def update(
        self,
        otel_name: Optional[str] = None,
        display_name: Optional[str] = None,
        aggregation: Optional[str | MetricAggregation] = None,
        enabled: Optional[bool] = None,
    ) -> None:
        """Update an OpenTelemetry metric configuration.

        .. versionadded:: v3.9

        Parameters
        ----------
        otel_name: Optional[str]
            Name of the reported metric.
        display_name: Optional[str]
            Name to display for the reported metric.
        aggregation: Optional[str]
            Aggregation type to use for the reported metric.
        enabled: Optional[bool]
            Whether the reported metric is shown in displays.
        """
        body: Dict[str, Any] = {}
        if otel_name:
            body["otelName"] = otel_name
        if display_name is not None:
            body["displayName"] = display_name
        if aggregation is not None:
            body["aggregation"] = str(aggregation)
        if enabled is not None:
            body["enabled"] = enabled

        if not self.id:
            raise ValueError("Cannot update a metric configuration without an ID")

        url = self._path.format(self.entity_type, self.entity_id) + self.id + "/"
        response = self._client.patch(url, json=body)

        # the response comes with the latest data, so update all the fields
        data = response.json()
        self.otel_name = data.get("otelName")
        self.display_name = data.get("displayName")
        self.aggregation = data.get("aggregation")
        self.enabled = data.get("enabled")

    def delete(self) -> None:
        """
        Delete this OpenTelemetry metric configuration.

        .. versionadded:: v3.9

        Returns
        -------
        None
        """

        if not self.id:
            raise ValueError("Cannot delete a metric configuration without an ID")

        url = self._path.format(self.entity_type, self.entity_id) + self.id + "/"
        self._client.delete(url)



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# limitations under the License.
from __future__ import annotations

import httpx

from datarobot.auth.exceptions import OAuthProviderNotFound


class OAuthServiceError(Exception):
    """
    The DR OAuth Providers Service request has failed
    """

    def __init__(
        self,
        message: str,
        status_code: int,
        provider_id: str | None = None,
        authorization_id: str | None = None,
    ) -> None:
        super().__init__(message)

        self.provider_id = provider_id
        self.authorization_id = authorization_id
        self.status_code = status_code


class OAuthServiceClientErr(OAuthServiceError):
    """
    The OAuth Providers Service request has failed due to a client error
    """

    def __init__(
        self,
        message: str,
        status_code: int = 422,
        provider_id: str | None = None,
        authorization_id: str | None = None,
    ) -> None:
        super().__init__(message, status_code, provider_id, authorization_id)


class OAuthServiceNotAuthenticated(OAuthServiceClientErr):
    """
    The OAuth Providers Service request has failed due to not being properly authenticated
    """

    def __init__(
        self,
        message: str,
        status_code: int = 401,
        provider_id: str | None = None,
        authorization_id: str | None = None,
    ) -> None:
        super().__init__(message, status_code, provider_id, authorization_id)


class OAuthServiceProviderNotFound(OAuthProviderNotFound, OAuthServiceClientErr):
    """
    The OAuth Providers Service doesn't recognize the provider ID.
    The provider could be accidentally deleted, never existed, or you don't have access to it.
    """

    def __init__(
        self,
        message: str,
        status_code: int = 404,
        provider_id: str | None = None,
        authorization_id: str | None = None,
    ) -> None:
        super().__init__(message, status_code, provider_id, authorization_id)


class OAuthAuthorizationNotFound(OAuthServiceClientErr):
    """
    The OAuth Providers Service could not find the authorization with the given ID.
    This can happen if the authorization was deleted by DataRobot User (there is a way to do that via DataRobot UI),
    never existed, or you don't have access to it.
    """

    def __init__(
        self,
        message: str,
        status_code: int = 404,
        provider_id: str | None = None,
        authorization_id: str | None = None,
    ) -> None:
        super().__init__(message, status_code, provider_id, authorization_id)


class OAuthAuthorizationExpired(OAuthServiceClientErr):
    """
    The OAuth Providers Service request has failed due to the authorization being expired.
    This can happen if
     - OAuth provider limits the lifetime of the refresh token
     - The user has revoked the authorization from this provider's account ahead of time
    """

    def __init__(
        self,
        message: str,
        status_code: int = 410,
        provider_id: str | None = None,
        authorization_id: str | None = None,
    ) -> None:
        super().__init__(message, status_code, provider_id, authorization_id)


class OAuthProviderUnavailable(OAuthServiceError):
    """
    The OAuth Providers Service request has failed due to the provider being unavailable
    """

    def __init__(
        self,
        message: str,
        status_code: int = 502,
        provider_id: str | None = None,
        authorization_id: str | None = None,
    ) -> None:
        super().__init__(message, status_code, provider_id, authorization_id)


class OAuthServiceUnavailable(OAuthServiceError):
    """
    The OAuth Providers Service request has failed due to the service being unavailable
    """

    def __init__(
        self,
        message: str,
        status_code: int = 500,
        provider_id: str | None = None,
        authorization_id: str | None = None,
    ) -> None:
        super().__init__(message, status_code, provider_id, authorization_id)


def raise_http_exception(
    resp: httpx.Response,
    provider_id: str | None = None,
    authorization_id: str | None = None,
) -> None:
    """
    Raise an exception based on the response status code.
    This is a placeholder function to handle errors in a more structured way.

    Args:
        resp (httpx.Response): The HTTP response object.
        provider_id (str | None): The ID of the OAuth provider, if applicable.
        authorization_id (str | None): The ID of the authorization, if applicable.

    Raises:
        OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
        OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
        OAuthServiceClientErr: For client errors (4xx status codes).
        OAuthServiceUnavailable: For server errors (5xx status codes).
        OAuthServiceError: For any other errors.
    """
    code = resp.status_code

    if code == 401:
        raise OAuthServiceNotAuthenticated(
            "Unauthorized access. Please check your DataRobot credentials.",
            provider_id=provider_id,
            authorization_id=authorization_id,
        )
    elif code == 502:
        raise OAuthProviderUnavailable(
            f"External OAuth provider is currently unavailable: {resp.text}",
            code,
            provider_id=provider_id,
            authorization_id=authorization_id,
        )
    elif 400 <= code < 500:
        raise OAuthServiceClientErr(
            f"Client error occurred: {resp.text}",
            code,
            provider_id=provider_id,
            authorization_id=authorization_id,
        )
    elif code >= 500:
        raise OAuthServiceUnavailable(
            f"The OAuth Providers Service or external OAuth provider is currently unavailable: {resp.text}",
            code,
            provider_id=provider_id,
            authorization_id=authorization_id,
        )
    else:
        raise OAuthServiceError(
            f"An error occurred while communicating with the OAuth provider: {resp.text}",
            code,
            provider_id=provider_id,
            authorization_id=authorization_id,
        )



#  Copyright 2025 DataRobot, Inc. and its affiliates.
#
#  All rights reserved.
#
#  DataRobot, Inc. Confidential.
#
#  This is unpublished proprietary source code of DataRobot, Inc.
#  and its affiliates.
#
#  The copyright notice above does not evidence any actual or intended
#  publication of such source code.
from __future__ import annotations

from datetime import datetime
from io import IOBase
import os
from typing import Dict, Iterable, List, Optional, Type, Union, cast

import dateutil
from requests import Response
import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import (
    DEFAULT_MAX_WAIT,
    DEFAULT_TIMEOUT,
    FileLocationType,
    FilesOverwriteStrategy,
    LocalSourceType,
    StrEnum,
)
from datarobot.errors import InvalidUsageError
from datarobot.models.api_object import APIObject
from datarobot.models.credential import CredentialDataSchema
from datarobot.models.dataset import _remove_empty_params
from datarobot.models.user_blueprints.models import HumanReadable
from datarobot.utils import assert_single_parameter
from datarobot.utils.pagination import unpaginate
from datarobot.utils.source import parse_source_type
from datarobot.utils.waiters import wait_for_async_resolution


class CatalogAccessType(StrEnum):
    """Enum of catalog access types for filtering search results."""

    OWNED = "owned"
    SHARED = "shared"
    CREATED = "created"
    ANY = "any"


_file_schema = t.Dict({
    t.Key("file_name") >> "name": String,
    t.Key("file_type") >> "type": String,
    t.Key("file_size") >> "size": Int(),
    t.Key("ingest_errors", optional=True): t.Or(String(allow_blank=True), t.Null),
})

_files_schema = t.Dict({
    t.Key("id"): String,
    t.Key("name"): String,
    t.Key("description", optional=True): t.Or(String, t.Null),
    t.Key("type"): String,
    t.Key("tags"): t.List(String),
    t.Key("num_files", optional=True): t.Int,
    t.Key("from_archive", optional=True): t.Bool(),
    t.Key("created_at"): t.Call(dateutil.parser.parse),
    t.Key("created_by", optional=True): t.Or(String, t.Null),
})

_file_catalog_search_schema = t.Dict({
    t.Key("id"): String,
    t.Key("catalog_name") >> "name": String,
    t.Key("description", optional=True): t.Or(String, t.Null),
    t.Key("info_creator_full_name") >> "created_by": String,
    t.Key("info_creation_date") >> "created_at": t.Call(dateutil.parser.parse),
    t.Key("last_modification_date") >> "last_modified_at": t.Call(dateutil.parser.parse),
    t.Key("last_modifier_full_name") >> "last_modified_by": String,
    t.Key("tags"): t.List(String),
    t.Key("num_files", optional=True): t.Or(Int, t.Null),
    t.Key("from_archive", optional=True): t.Or(bool, t.Null),
}).ignore_extra("*")


_files_stage_schema = t.Dict({"catalog_id": String, "stage_id": String}).ignore_extra("*")


class File(APIObject, HumanReadable):
    """
    Represents an individual file within a Files container.

    This class represents a single file contained within a Files container in the DataRobot catalog.
    It provides information about individual files such as name, size, and path within the archive.

    Attributes
    ----------
    name: str
        The name of the individual file.
    type: str
        The type of the file.
    size: int
        The size of the file in bytes.
    ingest_errors: str
        The errors encountered during ingestion of the file.
    """

    _converter = _file_schema.allow_extra("*")

    def __init__(
        self,
        name: str,
        type: str,
        size: int,
        ingest_errors: str | None = None,
    ):
        self.name = name
        self.type = type
        self.size = size
        self.ingest_errors = ingest_errors


class FilesStage(APIObject, HumanReadable):
    """A place to accumulate multiple uploaded files
    before they're added into corresponding files container.

    .. versionadded:: v3.10

    Attributes
    ----------
    catalog_id: str
        The unique identifier for the files container.
        The `FilesStage` can be applied only to that files container.
    stage_id: str
        The unique identifier for the `FilesStage` object.
    """

    _converter = _files_stage_schema

    def __init__(self, catalog_id: str, stage_id: str):
        self.catalog_id = catalog_id
        self.stage_id = stage_id

    def apply(self, overwrite: Optional[FilesOverwriteStrategy] = FilesOverwriteStrategy.RENAME) -> "Files":
        """
        Add the files uploaded into this `FilesStage` into the corresponding files container.
        You can call this method only once for a particular `FilesStage`.

        .. versionadded:: v3.10

        Parameters
        ----------
        overwrite: Optional[FilesOverwriteStrategy]
            How to deal with a name conflict between an existing file and an uploaded one.
            RENAME (default): rename an uploaded file using "<filename> (n).ext" pattern.'
            REPLACE: prefer an uploaded file.
            SKIP: prefer an existing file.'
            ERROR: return "HTTP 409 Conflict" response in case of a naming conflict. '

        Returns
        -------
        response: Files
            A fully armed and operational Files container
        """
        url = f"files/{self.catalog_id}/fromStage/"
        response = self._client.post(url, json={"stageId": self.stage_id, "overwrite": overwrite})
        return Files.get(response.json()["catalogId"])

    def upload(self, source: str | IOBase, file_name: str | None = None) -> None:
        """Upload a file into the `FilesStage`.

        .. versionadded:: v3.10

        Parameters
        ----------
        source: str | IOBase
            Local file path or a file-like object to upload.
        file_name: str
            The file name to apply on the server side.
        """
        url = f"files/{self.catalog_id}/stages/{self.stage_id}/upload/"

        if isinstance(source, str):
            fname = file_name or os.path.basename(source)
            self._client.build_request_with_file(
                method="post",
                url=url,
                fname=fname,
                file_path=source,
                read_timeout=DEFAULT_TIMEOUT.UPLOAD,
            )
        else:
            fname = cast(str, getattr(source, "name", file_name))
            self._client.build_request_with_file(
                method="post",
                url=url,
                fname=fname,
                filelike=source,
                read_timeout=DEFAULT_TIMEOUT.UPLOAD,
            )


class Files(APIObject):
    """
    Represents one or more files associated with a single entity in the DataRobot catalog.

    This class provides functionality to interact with files stored in the DataRobot catalog,
    including retrieving and updating file information and downloading file contents.

    .. versionadded:: v3.8

    Attributes
    ----------
    id: str
        The unique identifier for the files container.
    name: str
        The name of the files container.
    type: str
        The type of files container.
    tags: List[str]
        A list of tags associated with the files container.
    num_files: int
        The number of files in the container.
    from_archive: bool
        Whether the files container was extracted from an archive.
    created_at: datetime
        A timestamp from when the files container was created.
    created_by: str
        The username of the user who created the files container.
    description: Optional[str]
        An optional description of the files container.
    """

    _converter = _files_schema.allow_extra("*")
    _path = "files/"
    _async_status_location: str | None = None

    def __init__(
        self,
        id: str,
        name: str,
        type: str,
        tags: List[str],
        created_at: datetime,
        created_by: str,
        from_archive: bool = False,
        num_files: int | None = None,
        description: Optional[str] = None,
    ):
        self.id = id
        self.name = name
        self.type = type
        self.tags = tags
        self.num_files = num_files
        self.from_archive = from_archive
        self.created_at = created_at
        self.created_by = created_by
        self.description = description

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name!r}, id={self.id!r})"

    @classmethod
    def get(cls: Type["Files"], files_id: str) -> "Files":
        """
        Get information about a files container.

        .. versionadded:: v3.8

        Parameters
        ----------
        files_id: str
            The id of the files container.

        Returns
        -------
        files: Files
            The queried files container.
        """

        path = f"catalogItems/{files_id}/"
        return cls.from_location(path)

    def download(
        self,
        file_name: Optional[str] = None,
        file_path: Optional[str] = None,
        filelike: Optional[IOBase] = None,
        version_id: Optional[str] = None,
    ) -> None:
        """
        Retrieves uploaded file contents.
        Writes it to either the file or a file-like object that can write bytes.

        Only one of file_path or filelike can be provided. If a file-like object is
        provided, the user is responsible for closing it when they are done.

        The user must also have permission to download data.

        .. versionadded:: v3.8

        Parameters
        ----------
        file_name: Optional[str]
            The name of the file to download from the files container.
            If not provided, if the file container was created from an archive,
            it will download the original archive file.
            Otherwise, if there is only a single file contained in the container, it will download that.
        file_path: Optional[str]
            The destination to write the file to.
        filelike: Optional[IOBase]
            A file-like object to write to.  The object must be able to write bytes. The user is
            responsible for closing the object.
        version_id: Optional[str]
            If provided, download from the specified version instead of the latest version.
        Returns
        -------
        None
        """
        assert_single_parameter(("filelike", "file_path"), filelike, file_path)
        data = {}
        if file_name:
            data["fileName"] = file_name

        if version_id:
            path = f"{self._path}{self.id}/versions/{version_id}/downloads/"
        else:
            path = f"{self._path}{self.id}/downloads/"
        response = self._client.post(path, json=data, stream=True)
        if file_path:
            with open(file_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=1000):
                    f.write(chunk)
        if filelike:
            for chunk in response.iter_content(chunk_size=1000):
                filelike.write(chunk)

    @classmethod
    def upload(
        cls,
        source: Union[str, IOBase],
        tags: Optional[List[str]] = None,
        use_archive_contents: bool = True,
        *,
        wait_for_completion: bool = True,
    ) -> "Files":
        """
        This method covers Files container creation from local materials (file) and a URL.

        .. versionadded:: v3.8

        Parameters
        ----------
        source: str or file object
            Pass a URL, filepath, or file to create and return a Files container.
        tags: Optional[List[str]]
            A list of tags associated with the files container.
        use_archive_contents: bool (default: True)
            If True, extract archive contents and associate with the Files container.
            If False, the archive file will be uploaded as-is.

        Returns
        -------
        response: Files
            The Files container created from the uploaded data source.

        Raises
        ------
        InvalidUsageError
            If the source parameter cannot be determined to be a URL, filepath, or file object.

        Examples
        --------
        .. code-block:: python

            # Upload a local file
            files_a = Files.upload("./data/document.pdf")

            # Create a files container via URL with tags
            files_b = Files.upload(
                "https://example.com/document.pdf",
                tags=["web", "document"]
            )

            # Create files container using a local file object without extracting archive contents
            with open("./data/archive.zip", "rb") as file_pointer:
                files_c = Files.upload(file_pointer, use_archive_contents=False)
        """
        error_msg = f"Source parameter ({source}) cannot be determined to be a URL, filepath, or file object."
        try:
            source_type = parse_source_type(source)
        except InvalidUsageError:
            raise InvalidUsageError(error_msg)

        if source_type == FileLocationType.URL:
            return cls.create_from_url(
                url=cast(str, source),
                tags=tags,
                use_archive_contents=use_archive_contents,
                wait_for_completion=wait_for_completion,
            )
        elif source_type == FileLocationType.PATH:
            return cls.create_from_file(
                file_path=cast(str, source),
                tags=tags,
                use_archive_contents=use_archive_contents,
                wait_for_completion=wait_for_completion,
            )
        elif source_type == LocalSourceType.FILELIKE:
            return cls.create_from_file(
                filelike=cast(IOBase, source),
                tags=tags,
                use_archive_contents=use_archive_contents,
                wait_for_completion=wait_for_completion,
            )
        else:
            raise InvalidUsageError(error_msg)

    @classmethod
    def create_from_url(
        cls,
        url: str,
        tags: Optional[List[str]] = None,
        use_archive_contents: bool = True,
        max_wait: int = DEFAULT_MAX_WAIT,
        *,
        wait_for_completion: bool = True,
    ) -> "Files":
        """
        Create a new files container in the DataRobot catalog from a URL.

        This method uploads a file from a given URL to the DataRobot catalog. The method will wait
        for the upload to complete before returning.

        .. versionadded:: v3.8

        Parameters
        ----------
        url: str
            The URL of the file to upload. Must be accessible by the DataRobot server.
        tags: Optional[List[str]]
            A list of tags associated with the files container.
        use_archive_contents: bool
            If True, extract archive contents and associate with the Files container.
            If False, the archive file will be uploaded as-is. Defaults to True.
        max_wait: Optional[int]
            Maximum time in seconds to wait for the upload to complete. Defaults to DEFAULT_MAX_WAIT.

        Returns
        -------
        Files
            The newly created files container.

        Raises
        ------
        AsyncTimeoutError
            If the upload takes longer than max_wait seconds.
        """
        endpoint = f"{cls._path}fromURL/"
        payload: Dict[str, str] = {"url": url, "use_archive_contents": str(use_archive_contents)}

        response = cls._client.post(endpoint, data=payload)

        file = cls._get_files_from_async(response, wait_for_completion=wait_for_completion, max_wait=max_wait)

        if tags:
            file.modify(tags=tags)
        return file

    @classmethod
    def create_from_file(
        cls,
        file_path: Optional[str] = None,
        filelike: Optional[IOBase] = None,
        tags: Optional[List[str]] = None,
        use_archive_contents: bool = True,
        read_timeout: int = DEFAULT_TIMEOUT.UPLOAD,
        max_wait: int = DEFAULT_MAX_WAIT,
        *,
        wait_for_completion: bool = True,
    ) -> "Files":
        """
        A blocking call that creates a new files container from a file. Returns when the file has
        been successfully uploaded and processed.

        Warning: This function does not clean up its open files. If you pass a filelike, you are
        responsible for closing it. If you pass a file_path, this will create a file object from
        the file_path but will not close it.

        .. versionadded:: v3.8

        Parameters
        ----------
        file_path: Optional[str]
            The path to the file. This will create a files container object pointing to that file but
            will not close the local file.
        filelike: Optional[str]
            An open and readable file object.
        tags: Optional[List[str]]
            A list of tags associated with the files container.
        use_archive_contents: bool
            If True, extract archive contents and associate with the Files container.
            If False, the archive file will be uploaded as-is. Defaults to True.
        read_timeout: Optional[int]
            The maximum number of seconds to wait for the server to respond indicating that the
            initial upload is complete.
        max_wait: Optional[int]
            Time in seconds after which files container creation is considered unsuccessful.

        Returns
        -------
        response: Files
            A fully armed and operational Files container
        """
        assert_single_parameter(("filelike", "file_path"), file_path, filelike)

        upload_url = f"{cls._path}fromFile/"
        default_fname = "file"

        # Prepare additional form data
        form_data = {"use_archive_contents": str(use_archive_contents)}

        if file_path:
            fname = os.path.basename(file_path)
            response = cls._client.build_request_with_file(
                method="post",
                url=upload_url,
                fname=fname,
                file_path=file_path,
                read_timeout=read_timeout,
                form_data=form_data,
            )
        else:
            fname = getattr(filelike, "name", default_fname)
            response = cls._client.build_request_with_file(
                method="post",
                url=upload_url,
                fname=fname,
                filelike=filelike,
                read_timeout=read_timeout,
                form_data=form_data,
            )

        file = cls._get_files_from_async(response, wait_for_completion=wait_for_completion, max_wait=max_wait)

        if tags:
            file.modify(tags=tags)
        return file

    @classmethod
    def create_from_data_source(
        cls,
        data_source_id: str,
        tags: Optional[List[str]] = None,
        credential_id: Optional[str] = None,
        credential_data: Optional[Dict[str, str]] = None,
        use_archive_contents: bool = True,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> "Files":
        """
        A blocking call that creates a new Files container from data stored at a DataSource.
        Returns when the files container has been successfully uploaded and processed.

        .. versionadded:: v3.8

        Parameters
        ----------
        data_source_id: str
            The ID of the DataSource to use as the source of data.
        tags: Optional[List[str]]
            A list of tags associated with the files container.
        credential_id: Optional[str]
            The ID of the set of credentials to use for authentication.
        credential_data: Optional[Dict[str, str]]
            The credentials to authenticate with the database, to use instead of credential ID.
        use_archive_contents: bool
            If True, extract archive contents and associate with the files container.
            If False, the archive file will be uploaded as-is. Defaults to True.
        max_wait: Optional[int]
            Time in seconds after which files container creation is considered unsuccessful.

        Returns
        -------
        response: Files
            The Files container created from the uploaded data
        """
        base_data = {
            "data_source_id": data_source_id,
            "credential_id": credential_id,
            "credential_data": credential_data,
            "use_archive_contents": str(use_archive_contents),
        }
        data = _remove_empty_params(base_data)

        if "credential_data" in data:
            data["credential_data"] = CredentialDataSchema(data["credential_data"])

        upload_url = f"{cls._path}fromDataSource/"
        response = cls._client.post(upload_url, data=data)

        new_file_location = wait_for_async_resolution(cls._client, response.headers["Location"], max_wait)
        file = cls.from_location(new_file_location)
        if tags:
            file.modify(tags=tags)
        return file

    def update(self) -> None:
        """
        Updates the Files container attributes in place with the latest information from the server.

        .. versionadded:: v3.8

        Returns
        -------
        None
        """
        new_file = self.get(self.id)
        update_attrs = (
            "name",
            "description",
            "type",
            "tags",
            "num_files",
            "from_archive",
            "created_at",
            "created_by",
        )
        for attr in update_attrs:
            setattr(self, attr, getattr(new_file, attr))

    def modify(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ) -> None:
        """
        Modifies the Files container name, description and/or tags.
        Updates the object in place.

        .. versionadded:: v3.8

        Parameters
        ----------
        name: Optional[str]
            The new name of the files container.

        description: Optional[str]
            The new description of the files container.
        tags: Optional[List[str]]
            A list of tags attached to the files container.
            If any tags were previously specified for the files container, they will be overwritten.
            If omitted or None, keep previous tags.
            To clear them, specify [].

        Returns
        -------
        None
        """
        if name is None and description is None and tags is None:
            return
        # if we omit tags, it will clear them,
        # so in case where it is None, we want to preserve
        # categories and only clear them when it is []
        if tags is None:
            tags = self.tags

        url = f"catalogItems/{self.id}/"
        params = {"name": name, "description": description, "tags": tags}
        params = _remove_empty_params(params)

        response = self._client.patch(url, data=params)
        data = response.json()
        self.name = data.get("name", data.get("catalogName", ""))
        self.description = data["description"]
        self.tags = data["tags"]

    @classmethod
    def delete(cls, files_id: str) -> None:
        """
        Soft-deletes a files container. You cannot get, list, or do any actions with it, except for undeleting it.

        .. versionadded:: v3.8

        Parameters
        ----------
        files_id: str
            The id of the files container to mark for deletion.

        Returns
        -------
        None
        """
        cls._client.delete(f"{cls._path}{files_id}/")

    @classmethod
    def un_delete(cls, files_id: str) -> None:
        """
        Undeletes a previously deleted files container. If the files container was not deleted, nothing happens.

        .. versionadded:: v3.8

        Parameters
        ----------
        files_id: str
            The id of the files container to un-delete.

        Returns
        -------
        None
        """
        cls._client.patch(f"{cls._path}{files_id}/deleted/")

    @classmethod
    def search_catalog(
        cls,
        search: Optional[str] = None,
        tags: Optional[List[str]] = None,
        limit: int = 100,
        offset: int = 0,
        owner_user_id: Optional[str] = None,
        owner_username: Optional[str] = None,
        order_by: str = "-created",
        access_type: CatalogAccessType = CatalogAccessType.ANY,
    ) -> List["FilesCatalogSearch"]:
        """
        Fetch a list of the files container catalog entries the current user has access to
        based on an optional search term, tags, owner user info, or sort order.

        .. versionadded:: v3.8

        Parameters
        ----------
        search: Optional[str]
            A value to search for in the file's name, description, tags, etc. The search is case-insensitive.
            If no value is provided for this parameter, or if the empty string is used,
            or if the string contains only whitespace, no filtering will be done. Partial matching
            is performed on files container name and description fields while all other fields will only match
            if the search matches the whole value exactly.

        tags: Optional[List[str]]
            If provided, the results will be filtered to include only items with the specified tag.

        limit: Optional[int] (default: 100), at most this many results are returned. To specify no
            limit, use 0. The default may change and a maximum limit may be imposed without notice.

        offset: Optional[int] (default: 0), this many results will be skipped.

        owner_user_id: Optional[str]
            Filter results to those owned by one or more owner identified by UID.

        owner_username: Optional[str]
            Filter results to those owned by one or more owner identified by username.

        order_by: Optional[str] (default: '-created')
            Sort order which will be applied to a catalog list, valid options are "catalogName",
            "originalName", "description", "created", and "relevance". For all options other
            than relevance, you may prefix the attribute name with a dash to sort
            in descending order. E.g., orderBy='-catalogName'.

        access_type: Optional[CatalogAccessType] (default: CatalogAccessType.ANY)
            Access type used to filter returned results:
            'owned' items are owned by the requester, 'shared' items have been shared with the requester,
            'created' items have been created by the requester, and 'any' items matches all.

        Returns
        -------
        List[FilesCatalogSearch]
            A list of FileCatalogSearch objects matching the search criteria.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status
        datarobot.errors.ServerError
            If the server responded with 5xx status
        """
        return FilesCatalogSearch.search_catalog(
            search=search,
            tags=tags,
            limit=limit,
            offset=offset,
            owner_user_id=owner_user_id,
            owner_username=owner_username,
            order_by=order_by,
            access_type=access_type,
        )

    def clone(
        self,
        *,
        omit: str | List[str] | None = None,
        wait_for_completion: bool = True,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> Files:
        """Duplicate the files container.

        .. versionadded:: v3.10

        Parameters
        ----------
        omit:
            Don't duplicate some files.
        wait_for_completion:
            Set to *False* if you don't want to wait for the operation completion.
        max_wait:
            Raise AsyncTimeoutError if wait_for_completion=True
            and the operation took more than this number of seconds.
        """
        url = f"files/{self.id}/clone/"

        if isinstance(omit, str):
            omit = [omit]

        response = self._client.post(url, data={"omit": omit})
        return self._get_files_from_async(response, wait_for_completion=wait_for_completion, max_wait=max_wait)

    def create_stage(self) -> "FilesStage":
        """
        Create a new `FilesStage` for this files container.

        .. versionadded:: v3.10

        """
        response = self._client.post(f"files/{self.id}/stages/")
        return FilesStage.from_server_data(response.json())

    def apply_stage(self, stage: "FilesStage") -> None:
        """
        Apply the `FilesStage` for this files container.

        .. versionadded:: v3.10

        """
        file = stage.apply()
        self.num_files = file.num_files

    def copy(
        self,
        source_path: str | Iterable[str],
        *,
        target: str | None = None,
        target_files: Optional["Files"] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
        wait_for_completion: bool = True,
    ) -> "Files":
        """Copy file(s) and/or folder(s) within the same or into another files container.

        .. versionadded:: v3.10

        Parameters
        ----------
        source_path
            file(s) and/or folder(s) to copy.
        target
            Either a folder to copy file(s) into
            or a new file name if only one file is being copied.
        target_files
            Files collection to copy files into.
        max_wait
             Raise TimeoutError if the operation took more than this number of seconds to complete.
        """
        if isinstance(source_path, str):
            sources = [source_path]
        elif isinstance(source_path, Iterable):
            sources = list(source_path)
        else:
            raise ValueError(source_path)

        url = f"files/{self.id}/copyBatch/"
        response = self._client.post(
            url,
            json={
                "sources": sources,
                "target": target,
                "targetCatalogId": target_files and target_files.id,
            },
        )
        return self._get_files_from_async(response, max_wait=max_wait, wait_for_completion=wait_for_completion)

    def wait_for_completion(self) -> None:
        """Wait for initial upload completion."""
        if self._async_status_location is None:
            return

        location = wait_for_async_resolution(self._client, self._async_status_location)

        new_file = Files.from_location(location)
        self.num_files = new_file.num_files
        self.from_archive = new_file.from_archive
        self._async_status_location = None

    def list_contained_files(
        self,
        file_type: str = "",
        limit: int = 100,
        offset: int = 0,
        version_id: Optional[str] = None,
    ) -> List["File"]:
        """
        List all individual files within a Files container.

        .. versionadded:: v3.8

        This method retrieves information about all individual files contained within
        a Files object. This is useful for Files objects that contain multiple files
        or are archives.

        Parameters
        ----------
        file_type: str
            Filter results by file type (e.g., 'txt', 'pdf').
        limit: Optional[int] (default: 100)
            Maximum number of files to return. Set to 0 for no limit.
        offset: Optional[int] (default: 0)
            Number of files to skip before returning results.
        version_id: Optional[str]
            If provided, retrieve from the specified version instead of the latest version.


        Returns
        -------
        List[File]
            A list of File objects representing individual files within the Files container.

        Raises
        ------
        ClientError
            If the Files object is not found or access is denied.
        ServerError
            If there's a server-side error while retrieving the file list.
        """
        if version_id:
            endpoint = f"{self._path}{self.id}/versions/{version_id}/allFiles/"
        else:
            endpoint = f"{self._path}{self.id}/allFiles/"
        params: Dict[str, Union[int, str]] = {"offset": offset}
        if file_type:
            params["file_type"] = file_type
        if limit == 0:
            files_data = list(unpaginate(endpoint, params, self._client))
        else:
            params["limit"] = limit
            files_data = self._client.get(endpoint, params=params).json()["data"]

        return [File.from_server_data(file_data) for file_data in files_data]

    @classmethod
    def _get_files_from_async(cls, response: Response, *, wait_for_completion: bool = True, max_wait: int) -> "Files":
        """Get `Files` entity from the response.
        Conditionally wait for an async operation resolution.

        Parameters
        ----------
        response
            HTTP response for an action which has just created a new files container.
        wait_for_completion
            Set the parameter to False to get the entity before the operation completed.
        max_wait:
            Raise AsyncTimeoutError if wait_for_completion=True
            and the operation took more than this number of seconds.
        """
        if wait_for_completion:
            new_file_location = wait_for_async_resolution(cls._client, response.headers["Location"], max_wait)
            return cls.from_location(new_file_location)

        else:
            entity = cls.from_location(f"catalogItems/{response.json()['catalogId']}/")
            if "Location" in response.headers:
                entity.set_async_status_location(response.headers["Location"])

            return entity

    def set_async_status_location(self, async_status_location: str) -> None:
        """Assign a URL to keep track of an async operation completion."""
        self._async_status_location = async_status_location


class FilesCatalogSearch(APIObject, HumanReadable):
    """
    An APIObject representing a file catalog entry the current user has access to
    based on an optional search term and/or tags.

    .. versionadded:: v3.8

    Attributes
    ----------
    id: str
        The ID of the catalog entry linked to the file.

    name: str
        The name of the files container in the catalog.

    description: Optional[str] (default: None)
        The description of the file.

    tags: List[str]
        A list of tags associated with the files container.

    num_files: Optional[int] (default: None)
        The number of files in the container.

    from_archive: Optional[bool] (default: None)
        Whether the files container was extracted from an archive.

    created_by: str
        The name of the user that created the file.

    created_at: datetime
        The timestamp when the catalog item was created.

    last_modified_by: str
        The name of the user that last modified the catalog item.

    last_modified_at: datetime
        The timestamp when the catalog item was last modified.
    """

    _path = "catalogItems/"
    _converter = _file_catalog_search_schema

    def __init__(
        self,
        id: str,
        name: str,
        created_by: str,
        created_at: datetime,
        last_modified_at: datetime,
        last_modified_by: str,
        tags: List[str],
        from_archive: Optional[bool] = None,
        num_files: Optional[int] = None,
        description: Optional[str] = None,
    ) -> None:
        self.id = id
        self.name = name
        self.created_by = created_by
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        self.last_modified_by = last_modified_by
        self.description = description
        self.tags = tags
        self.num_files = num_files
        self.from_archive = from_archive

    @classmethod
    def search_catalog(
        cls,
        search: Optional[str] = None,
        tags: Optional[List[str]] = None,
        limit: int = 100,
        offset: int = 0,
        owner_user_id: Optional[str] = None,
        owner_username: Optional[str] = None,
        order_by: str = "-created",
        access_type: CatalogAccessType = CatalogAccessType.ANY,
    ) -> List["FilesCatalogSearch"]:
        """
        Fetch a list of the files container entries the current user has access to
        based on an optional search term, tags, owner user info, or sort order.

        .. versionadded:: v3.8

        Parameters
        ----------
        search: Optional[str]
            A value to search for in the file's name, description, tags, etc. The search is case-insensitive.
            If no value is provided for this parameter, or if the empty string is used,
            or if the string contains only whitespace, no filtering will be done. Partial matching
            is performed on files container name and description fields while all other fields will only match
            if the search matches the whole value exactly.

        tags: Optional[List[str]]
            If provided, the results will be filtered to include only items with the specified tag.

        limit: Optional[int] (default: 100), at most this many results are returned. To specify no
            limit, use 0. The default may change and a maximum limit may be imposed without notice.

        offset: Optional[int] (default: 0), this many results will be skipped.

        owner_user_id: Optional[str]
            Filter results to those owned by one or more owner identified by UID.

        owner_username: Optional[str]
            Filter results to those owned by one or more owner identified by username.

        order_by: Optional[str] (default: '-created')
            Sort order which will be applied to a catalog list, valid options are "catalogName",
            "originalName", "description", "created", and "relevance". For all options other
            than relevance, you may prefix the attribute name with a dash to sort
            in descending order. e.g., orderBy='-catalogName'.

        access_type: Optional[CatalogAccessType] (default: CatalogAccessType.ANY)
            Access type used to filter returned results:
            'owned' items are owned by the requester, 'shared' items have been shared with the requester,
            'created' items have been created by the requester, and 'any' items matches all.

        Returns
        -------
        List[FilesCatalogSearch]
            A list of FileCatalogSearch objects matching the search criteria.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        params = {
            "search_for": search,
            "tag": [tag.strip() for tag in tags if tag] if tags else None,
            "type": "files",
            "limit": limit,
            "offset": offset,
            "owner_user_id": owner_user_id,
            "owner_username": owner_username,
            "order_by": order_by,
            "access_type": access_type,
        }
        params = _remove_empty_params(params)
        if limit == 0:
            data = list(unpaginate(cls._path, params, cls._client))
        else:
            data = cls._client.get(cls._path, params=params).json()["data"]

        return [cls.from_server_data(a) for a in data]



# flake8: noqa

from .logs import OtelLogEntry
from .metric_config import OtelMetricConfig
from .metric_enums import MetricAggregation, MetricResolution
from .metric_summary import OtelMetricSummary
from .metric_values import OtelMetricBucketValue, OtelMetricValue



# flake8: noqa
from .intervention import GuardInterventionCondition
from .configuration import ModerationConfiguration
from .model_version_update import ModelVersionUpdate
from .overall import OverallModerationConfig
from .template import ModerationTemplate
